/*
 * Copyright 2018 The CovenantSQL Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package main

import (
	"flag"
	"fmt"
	"math/rand"
	"net"
	"net/http"
	_ "net/http/pprof"
	"os"
	"os/signal"
	"runtime"
	//"runtime/trace"
	"syscall"
	"time"

	"github.com/CovenantSQL/CovenantSQL/conf"
	"github.com/CovenantSQL/CovenantSQL/crypto/asymmetric"
	"github.com/CovenantSQL/CovenantSQL/crypto/kms"
	"github.com/CovenantSQL/CovenantSQL/metric"
	"github.com/CovenantSQL/CovenantSQL/route"
	"github.com/CovenantSQL/CovenantSQL/rpc"
	"github.com/CovenantSQL/CovenantSQL/utils"
	"github.com/CovenantSQL/CovenantSQL/utils/log"
	"github.com/CovenantSQL/CovenantSQL/worker"
	"github.com/cyberdelia/go-metrics-graphite"
	"github.com/rcrowley/go-metrics"
)

const logo = `
   ______                                  __  _____ ____    __ 
  / ____/___ _   _____  ____  ____ _____  / /_/ ___// __ \  / / 
 / /   / __ \ | / / _ \/ __ \/ __  / __ \/ __/\__ \/ / / / / /
/ /___/ /_/ / |/ /  __/ / / / /_/ / / / / /_ ___/ / /_/ / / /___
\____/\____/|___/\___/_/ /_/\__,_/_/ /_/\__//____/\___\_\/_____/

   _____ ______   ___  ________   _______   ________     
  |\   _ \  _   \|\  \|\   ___  \|\  ___ \ |\   __  \    
  \ \  \\\__\ \  \ \  \ \  \\ \  \ \   __/|\ \  \|\  \   
   \ \  \\|__| \  \ \  \ \  \\ \  \ \  \_|/_\ \   _  _\  
    \ \  \    \ \  \ \  \ \  \\ \  \ \  \_|\ \ \  \\  \| 
     \ \__\    \ \__\ \__\ \__\\ \__\ \_______\ \__\\ _\ 
      \|__|     \|__|\|__|\|__| \|__|\|_______|\|__|\|__|
                                                       
`

var (
	version = "unknown"
)

var (
	// config
	configFile string
	genKeyPair bool
	metricLog  bool

	// profile
	cpuProfile     string
	memProfile     string
	profileServer  string
	metricGraphite string
	traceFile      string

	// other
	noLogo      bool
	showVersion bool
)

const name = `cql-minerd`
const desc = `CovenantSQL is a Distributed Database running on BlockChain`

func init() {
	flag.BoolVar(&noLogo, "nologo", false, "Do not print logo")
	flag.BoolVar(&metricLog, "metricLog", false, "Print metrics in log")
	flag.BoolVar(&showVersion, "version", false, "Show version information and exit")
	flag.BoolVar(&genKeyPair, "genKeyPair", false, "Gen new key pair when no private key found")
	flag.BoolVar(&asymmetric.BypassSignature, "bypassSignature", false,
		"Disable signature sign and verify, for testing")

	flag.StringVar(&configFile, "config", "./config.yaml", "Config file path")

	flag.StringVar(&profileServer, "profileServer", "", "Profile server address, default not started")
	flag.StringVar(&cpuProfile, "cpu-profile", "", "Path to file for CPU profiling information")
	flag.StringVar(&memProfile, "mem-profile", "", "Path to file for memory profiling information")
	flag.StringVar(&metricGraphite, "metricGraphiteServer", "", "Metric graphite server to push metrics")
	flag.StringVar(&traceFile, "traceFile", "", "trace profile")

	flag.Usage = func() {
		fmt.Fprintf(os.Stderr, "\n%s\n\n", desc)
		fmt.Fprintf(os.Stderr, "Usage: %s [arguments]\n", name)
		flag.PrintDefaults()
	}
}

func initLogs() {
	log.Infof("%#v starting, version %#v", name, version)
	log.Infof("%#v, target architecture is %#v, operating system target is %#v", runtime.Version(), runtime.GOARCH, runtime.GOOS)
	log.Infof("role: %#v", conf.RoleTag)
}

func main() {
	// set random
	rand.Seed(time.Now().UnixNano())
	log.SetLevel(log.InfoLevel)
	flag.Parse()

	if showVersion {
		fmt.Printf("%v %v %v %v %v\n",
			name, version, runtime.GOOS, runtime.GOARCH, runtime.Version())
		os.Exit(0)
	}

	flag.Visit(func(f *flag.Flag) {
		log.Infof("Args %#v : %s", f.Name, f.Value)
	})

	var err error
	conf.GConf, err = conf.LoadConfig(configFile)
	if err != nil {
		log.WithField("config", configFile).WithError(err).Fatal("load config failed")
	}

	if conf.GConf.Miner == nil {
		log.Fatal("miner config does not exists")
	}
	if conf.GConf.Miner.MetricCollectInterval.Seconds() <= 0 {
		log.Fatal("miner metric collect interval is invalid")
	}
	if conf.GConf.Miner.MaxReqTimeGap.Seconds() <= 0 {
		log.Fatal("miner request time gap is invalid")
	}

	kms.InitBP()
	log.Debugf("config:\n%#v", conf.GConf)

	// init log
	initLogs()

	if !noLogo {
		fmt.Print(logo)
	}

	// init profile, if cpuProfile, memProfile length is 0, nothing will be done
	utils.StartProfile(cpuProfile, memProfile)

	// set generate key pair config
	conf.GConf.GenerateKeyPair = genKeyPair

	// start rpc
	var server *rpc.Server
	if server, err = initNode(); err != nil {
		log.WithError(err).Fatal("init node failed")
	}

	if conf.GConf.Miner.IsTestMode {
		// miner test mode enabled
		log.Debug("miner test mode enabled")
	}

	// stop channel for all daemon routines
	stopCh := make(chan struct{})
	defer close(stopCh)

	if len(profileServer) > 0 {
		go func() {
			log.Println(http.ListenAndServe(profileServer, nil))
		}()
	}

	// start metric collector
	go func() {
		mc := metric.NewCollectClient()
		tick := time.NewTicker(conf.GConf.Miner.MetricCollectInterval)
		defer tick.Stop()

		for {
			// if in test mode, upload metrics to all block producer
			if conf.GConf.Miner.IsTestMode {
				// upload to all block producer
				for _, bpNodeID := range route.GetBPs() {
					mc.UploadMetrics(bpNodeID)
				}
			} else {
				// choose block producer
				if bpID, err := rpc.GetCurrentBP(); err != nil {
					log.Error(err)
					continue
				} else {
					mc.UploadMetrics(bpID)
				}
			}

			select {
			case <-stopCh:
				return
			case <-tick.C:
			}
		}
	}()

	// start dbms
	var dbms *worker.DBMS
	if dbms, err = startDBMS(server); err != nil {
		log.WithError(err).Fatal("start dbms failed")
	}

	defer dbms.Shutdown()

	// start rpc server
	go func() {
		server.Serve()
	}()
	defer func() {
		server.Listener.Close()
		server.Stop()
	}()

	signalCh := make(chan os.Signal, 1)
	signal.Notify(
		signalCh,
		syscall.SIGINT,
		syscall.SIGTERM,
	)
	signal.Ignore(syscall.SIGHUP, syscall.SIGTTIN, syscall.SIGTTOU)

	if metricLog {
		go metrics.Log(metrics.DefaultRegistry, 5*time.Second, log.StandardLogger())
	}

	if metricGraphite != "" {
		addr, err := net.ResolveTCPAddr("tcp", metricGraphite)
		if err != nil {
			log.WithError(err).Error("resolve metric graphite server addr failed")
			return
		}
		minerName := fmt.Sprintf("miner-%s", conf.GConf.ThisNodeID[len(conf.GConf.ThisNodeID)-5:])
		go graphite.Graphite(metrics.DefaultRegistry, 5*time.Second, minerName, addr)
	}

	//if traceFile != "" {
	//	f, err := os.Create(traceFile)
	//	if err != nil {
	//		log.WithError(err).Fatal("failed to create trace output file")
	//	}
	//	defer func() {
	//		if err := f.Close(); err != nil {
	//			log.WithError(err).Fatal("failed to close trace file")
	//		}
	//	}()

	//	if err := trace.Start(f); err != nil {
	//		log.WithError(err).Fatal("failed to start trace")
	//	}
	//	defer trace.Stop()
	//}

	<-signalCh
	utils.StopProfile()

	log.Info("miner stopped")
}
